﻿/*
 
 Attrice Corporation 2008-2009

 This source code and the compiled software are provided "AS IS", without warranty of any kind, express or implied, 
 including but not limited to the warranties of merchantability, fitness for a particular purpose 
 and noninfringement. In no event shall the authors or copyright holders be liable for any claim, 
 damages or other liability, whether in an action of contract, tort or otherwise, arising from, 
 out of or in connection with the software or the use or other dealings in the software. 

 Questions? Contact us at info@attrice.info
  
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.IO;
using System.Xml;
using System.Diagnostics;
using System.Windows.Forms;

using EnvDTE;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.StyleCop;

using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;

namespace Attrice.TeamFoundation.StyleCop
{
  [Serializable]
  public class Policy : PolicyBase, IVsUpdateSolutionEvents
  {
    public override string Type
    {
      get { return "C# StyleCop"; }
    }

    public override string TypeDescription
    {
      get { return "C# StyleCop integrated check-in policy"; }
    }

    public override string Description
    {
      get { return "C# StyleCop integrated check-in policy"; }
    }

    /// <summary>
    /// Instance of VS IDE to work with current solution
    /// </summary>
    [NonSerialized]
    private _DTE _dte;

    private const string ErrorFilePrefix = "In file: ";
    private const string ErrorFileLinePrefix = "; line ";

    /// <summary>
    /// List of errors after policy evaluation; need this
    /// to keep a reference to the error (provider does not do that)
    /// </summary>
    [NonSerialized]
    private List<ErrorTask> _errorTasks;

    /// <summary>
    /// Error provider used to log errors into VS
    /// </summary>
    [NonSerialized]
    private ErrorListProvider _errorsProvider;

    /// <summary>
    /// Cookie for solution build events sink
    /// </summary>
    [NonSerialized]
    private uint _solutionBuildEventsCookie;

    /// <summary>
    /// Provider identifier
    /// </summary>
    private static readonly Guid providerGuid = new Guid("{6C9AA2B0-859B-4477-A39C-B85195611A7E}");

    public override void Initialize(IPendingCheckin pendingCheckin)
    {
      Trace.TraceInformation("Initialize");
      base.Initialize(pendingCheckin);
      PendingCheckin.PendingChanges.CheckedPendingChangesChanged += new EventHandler(PendingChangesChanged);
      _dte = pendingCheckin.GetService(typeof(_DTE)) as _DTE;

      // initialize error provider
      if (_errorsProvider != null)
      {
        _errorsProvider.Tasks.Clear();
      }
      else
      {
        IServiceProvider serviceProvider = new ServiceProvider(_dte.Application as Microsoft.VisualStudio.OLE.Interop.IServiceProvider);
        Debug.Assert(serviceProvider != null);
        _errorsProvider = new ErrorListProvider(serviceProvider);
        _errorsProvider.ProviderName = "StyleCop checkin policy";
        _errorsProvider.ProviderGuid = providerGuid;
      }
    }

    public override void Dispose()
    {
      Trace.TraceInformation("Dispose");
      // if have subscribed to build events, try to unsubscribe
      if (_solutionBuildEventsCookie != 0)
      {
        IVsSolutionBuildManager buildManager = Microsoft.VisualStudio.Shell.Package.GetGlobalService(typeof(IVsSolutionBuildManager)) as IVsSolutionBuildManager;
        buildManager.UnadviseUpdateSolutionEvents(_solutionBuildEventsCookie);
        _solutionBuildEventsCookie = 0;
      }
      // if have registered errors provider, clean it up
      if (_errorsProvider != null)
      {
        _errorsProvider.Tasks.Clear();
        _errorsProvider.Dispose();
        _errorsProvider = null;
      }
      // if have list of tasks, clear it out
      if (_errorTasks != null)
      {
        _errorTasks.Clear();
        _errorTasks = null;
      }
      if (_violations != null)
      {
        _violations.Clear();
        _violations = null;
      }
      if (_fileProjectMap != null)
      {
        _fileProjectMap.Clear();
        _fileProjectMap = null;
      }
      if (_externalFiles != null)
      {
        _externalFiles.Clear();
        _externalFiles = null;
      }

      PendingCheckin.PendingChanges.CheckedPendingChangesChanged -= new EventHandler(PendingChangesChanged);
      base.Dispose();
    }

    private void PendingChangesChanged(object sender, EventArgs e)
    {
      Trace.TraceInformation("PendingChangesChanged");
      //if (!Disposed)
      //    EvaluateAgain();
    }

    private void EvaluateAgain()
    {
      PolicyFailure[] failures = Evaluate();
      OnPolicyStateChanged(failures);
    }

    /// <summary>
    /// Called when double clicking on policy violation. Shows details dialog
    /// </summary>
    /// <param name="failure"></param>
    public override void Activate(PolicyFailure failure)
    {
      if (failure.Policy == this && _violations != null && _violations.Count > 0)
      {
        // try to navigate to file position using error task
        if (NavigateUsingTask(failure.Message))
        {
          return;
        }
        // if could not navigate to file, show custom form
        using (Form detailsDialog = new DetailsForm(_violations))
          detailsDialog.ShowDialog();
      }
      else if (failure.Policy == this && _externalFiles != null && _externalFiles.Count > 0)
        using (Form filesDialog = new ExternalFilesForm(_externalFiles))
          filesDialog.ShowDialog();
      else
        base.Activate(failure);
    }

    private bool NavigateUsingTask(string message)
    {
      // get file name and line using message formatting
      // then find appropriate ErrorTask and navigate using it
      if (message.Contains(ErrorFilePrefix))
      {
        int indexStart = message.IndexOf(ErrorFilePrefix) + ErrorFilePrefix.Length;
        int indexEnd = message.IndexOf(ErrorFileLinePrefix, indexStart);
        string fileName = message.Substring(indexStart, indexEnd - indexStart);
        indexEnd += ErrorFileLinePrefix.Length;
        int lineNumber = Int32.Parse(message.Substring(indexEnd));
        foreach (ErrorTask task in _errorTasks)
        {
          if (task.Document.Equals(fileName, StringComparison.OrdinalIgnoreCase)
            && task.Line + 1 == lineNumber)
          {
            ErrorTaskNavigate(task, EventArgs.Empty);
            return true;
          }
        }
      }
      return false;
    }

    /// <summary>
    /// Holds XML with source analysis configuration
    /// </summary>
    private string _sourceCodeAnalysisSettings;

    /// <summary>
    /// Load settings from external file into file specified as input
    /// </summary>
    /// <param name="temporarySettingsFilePath">Target file</param>
    /// <returns>True if external settings imported into target file, false o/w</returns>
    private bool LoadExternalSettings(string temporarySettingsFilePath)
    {
      while (MessageBox.Show("Do you want to import existing settings?", "StyleCop check-in policy", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
      {
        using (OpenFileDialog dialog = new OpenFileDialog())
        {
          dialog.Filter = "StyleCop settings (*.stylecop)|*.stylecop|All files (*.*)|*.*";
          dialog.Multiselect = false;
          dialog.Title = "Select StyleCop settings file";
          if (dialog.ShowDialog() == DialogResult.OK)
          {
            using (StreamReader reader = File.OpenText(dialog.FileName))
            {
              using (StreamWriter writer = File.CreateText(temporarySettingsFilePath))
              {
                writer.Write(reader.ReadToEnd());
                writer.Close();
              }
            }
            Trace.TraceInformation("Loading StyleCop settings from {0}", dialog.FileName);
            return true;
          }
        }
      }
      return false;
    }

    /// <summary>
    /// Called when policy is created or its definitions edited
    /// </summary>
    /// <param name="policyEditArgs"></param>
    /// <returns></returns>
    public override bool Edit(IPolicyEditArgs policyEditArgs)
    {
      Trace.TraceInformation("Edit IsNew {0}", policyEditArgs.IsNew);
      try
      {
        DateTime settingsFileDate = DateTime.MinValue;
        bool externalExport = false;
        string settingsFile = Path.GetTempFileName();
        if (policyEditArgs.IsNew)
        {
          // user may select file to import settings from
          if (LoadExternalSettings(settingsFile))
          {
            externalExport = true;
            // remember file change date (need later to handle issue with
            // configuration dialog)
            settingsFileDate = new FileInfo(settingsFile).LastWriteTime;
            Trace.TraceInformation("Edit for new policy; imported settings into {0}", settingsFile);
          }
          else
          {
            // if new policy, provide path, the dialog will create file
            Trace.TraceInformation("Edit for new policy; file will be created at {0}", settingsFile);
            DeleteFile(settingsFile);
          }
        }
        else
        {
          Debug.Assert(_sourceCodeAnalysisSettings != null);
          // if existing policy, create temp definitions file
          Trace.TraceInformation("Edit existing policy; edit settings file at {0}", settingsFile);
          using (StreamWriter writer = File.CreateText(settingsFile))
          {
            writer.Write(_sourceCodeAnalysisSettings);
            writer.Close();
          }
        }
        // run source analysis UI dialog
        StyleCopCore core = new StyleCopCore(null, null);
        core.Initialize(null, true);
        core.WriteResultsCache = false;
        core.DisplayUI = true;
        try
        {
          // HACK!
          // if not loaded from external settings,file creation 
          // date is approximately date of ShowSettings call (+/- 100ms)
          if (!externalExport)
            settingsFileDate = DateTime.Now;
          if (!core.ShowSettings(settingsFile))
          {
            if (policyEditArgs.IsNew)
            {
              // there is issue with the dialog
              // 1. File is always created
              // 2. False is returned if user did not change anything and hit OK
              // 3. False is returned on Cancel
              // So the file is tested - if it was changed after UI dialog is shown
              // then user hit OK, o/w Cancel 
              DateTime settingsFileDateNow = new FileInfo(settingsFile).LastWriteTime;
              if (settingsFileDate != DateTime.MinValue
                  && (settingsFileDateNow - settingsFileDate).TotalMilliseconds < 100)
                return false;
            }
            else
              return false;
          }
          // read settings into policy (the dialog saved changes to file)                
          using (StreamReader reader = new StreamReader(File.OpenRead(settingsFile)))
          {
            _sourceCodeAnalysisSettings = reader.ReadToEnd();
            reader.Close();
          }
          return true;
        }
        finally
        {
          DeleteFile(settingsFile);
        }
      }
      catch (Exception ex)
      {
        Trace.TraceError("Exception in Edit {0} {1}", ex.Message, ex.StackTrace);
        throw ex;
      }
    }

    private void ProjectSettingsChanged(object sender, EventArgs e)
    {
      throw new NotImplementedException();
    }

    /// <summary>
    /// We support editing the policy
    /// </summary>
    public override bool CanEdit
    {
      get { return true; }
    }

    public override string InstallationInstructions
    {
      get { return "Please refer to policy documentation for installation instructions."; }
    }

    /// <summary>
    /// Recursively iterate over all project items
    /// </summary>
    /// <param name="projectItems"></param>
    /// <param name="projectFiles"></param>
    private static void GetProjectFilesRecursively(ProjectItems projectItems, Dictionary<string, string> projectFiles)
    {
      if (projectItems == null)
        return;
      if (projectItems.Count == 0)
        return;
      for (int i = 1; i < projectItems.Count + 1; i++)
      {
        // get files contained
        ProjectItem item = projectItems.Item(i);
        for (short j = 1; j < item.FileCount + 1; j++)
        {
          string fileName = item.get_FileNames(j);
          Trace.TraceInformation("GetProjectFilesRecursively file {0} {2}, {1}", j, fileName, item.Kind);
          // filter out weird cases - folders in Web site & files appearing twice
          if (File.Exists(fileName) && !projectFiles.ContainsKey(fileName))
            projectFiles.Add(fileName, item.Kind);
        }
        // get child items
        GetProjectFilesRecursively(item.ProjectItems, projectFiles);
        // get sub project
        if (item.SubProject != null)
        {
          Trace.TraceInformation("GetProjectFilesRecursively sub-project {0}", item.get_FileNames(1));
          GetProjectFilesRecursively(item.SubProject.ProjectItems, projectFiles);
        }
      }
    }

    /// <summary>
    /// Start recursive iteration on project to get all contained files
    /// </summary>
    /// <param name="project"></param>
    /// <returns></returns>
    private static Dictionary<string, string> GetProjectFiles(ProjectItems projectItems)
    {
      Dictionary<string, string> projectFiles = new Dictionary<string, string>(StringComparer.CurrentCultureIgnoreCase);
      GetProjectFilesRecursively(projectItems, projectFiles);
      return projectFiles;
    }

    /// <summary>
    /// Given data structure with projects/list of files find corresponding project for
    /// given file
    /// </summary>
    /// <param name="projects"></param>
    /// <param name="file"></param>
    /// <returns></returns>
    private Project FindProject(Dictionary<Project, Dictionary<string, string>> projects, string file)
    {
      foreach (Project project in projects.Keys)
      {
        Dictionary<string, string> files = projects[project];
        if (files.ContainsKey(file))
          return project;
      }
      return null;
    }

    /// <summary>
    /// Add all project items list into dictionary 
    /// </summary>
    /// <param name="project"></param>
    /// <param name="projects"></param>
    private void IterateProjectItems(Project project, Dictionary<Project, Dictionary<string, string>> projects)
    {
      Trace.TraceInformation("IterateProjectItems '{0}':'{1}':{2}", project.Name, project.Kind, project.FileName);
      if (!IsWebServiceProject(project.Kind)
          && !IsWebProject(project.Kind)
          && !IsCSharpProject(project.Kind))
        return;
      // check if the language is C#
      if (project.CodeModel == null || String.Compare("{B5E9BD34-6D3E-4B5D-925E-8A43B79820B4}", project.CodeModel.Language, true) != 0)
        return;
      projects.Add(project, GetProjectFiles(project.ProjectItems));
    }

    /// <summary>
    /// Iterate over sub-projects (needed for solution folder)s
    /// </summary>
    /// <param name="project"></param>
    /// <param name="projects"></param>
    private void IterateSubProjects(Project project, Dictionary<Project, Dictionary<string, string>> projects)
    {
      Trace.TraceInformation("IterateSubProjects '{0}':'{1}':{2}", project.Name, project.Kind, project.FileName);
      if (project.ProjectItems == null)
        return;
      for (int i = 1; i < project.ProjectItems.Count + 1; i++)
      {
        ProjectItem item = project.ProjectItems.Item(i);
        // solution sub-folder
        if (!IsSolutionSubFolder(item.Kind))
          continue;

        if (item.SubProject != null)
        {
          // if subfolder - drill down
          if (IsSolutionFolder(item.SubProject.Kind))
            IterateSubProjects(item.SubProject, projects);
          else // load project items
            IterateProjectItems(item.SubProject, projects);
        }
      }
    }

    #region Static project Guid checkers
    /// <summary>
    /// Checks if the kind GUID is that of solution subfolder
    /// </summary>
    /// <param name="kind"></param>
    /// <returns></returns>
    private static bool IsSolutionSubFolder(string kind)
    {
      Debug.Assert(kind != null);
      return kind.Equals("{66A26722-8FB5-11D2-AA7E-00C04F688DDE}", StringComparison.OrdinalIgnoreCase);
    }

    /// <summary>
    /// Checks if the kind GUID is that of solution folder
    /// </summary>
    /// <param name="kind"></param>
    /// <returns></returns>
    private static bool IsSolutionFolder(string kind)
    {
      Debug.Assert(kind != null);
      return kind.Equals("{66A26720-8FB5-11D2-AA7E-00C04F688DDE}", StringComparison.OrdinalIgnoreCase);
    }

    /// <summary>
    /// Checks if the kind GUID is that of web site project
    /// </summary>
    /// <param name="kind"></param>
    /// <returns></returns>
    private static bool IsWebServiceProject(string kind)
    {
      Debug.Assert(kind != null);
      return kind.Equals("{1DF4D691-79FD-4E3E-B7A6-A1A5DC772BBD}", StringComparison.OrdinalIgnoreCase);
    }

    /// <summary>
    /// Checks if the kind GUID is that of web site project
    /// </summary>
    /// <param name="kind"></param>
    /// <returns></returns>
    private static bool IsWebProject(string kind)
    {
      Debug.Assert(kind != null);
      return kind.Equals("{E24C65DC-7377-472b-9ABA-BC803B73C61A}", StringComparison.OrdinalIgnoreCase);
    }

    /// <summary>
    /// Checks if the kind GUID is that of C# project
    /// </summary>
    /// <param name="kind"></param>
    /// <returns></returns>
    private static bool IsCSharpProject(string kind)
    {
      Debug.Assert(kind != null);
      return kind.Equals("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}", StringComparison.OrdinalIgnoreCase);
    }
    #endregion Static project Guid checkers

    public override PolicyFailure[] Evaluate()
    {
      _violations = new Dictionary<string, List<ViolationEventArgs>>(StringComparer.CurrentCultureIgnoreCase);
      _fileProjectMap = new Dictionary<string, Project>(StringComparer.CurrentCultureIgnoreCase);
      _externalFiles = new List<string>();

      try
      {
        // check if there is solution (if none, fail since cannot figure out projects)
        if (_dte.Solution == null || _dte.Solution.Projects.Count == 0)
          return new PolicyFailure[] { new PolicyFailure("Source Analysis policy cannot be evaluated with empty or no solution.", this) };

        // get list of solution projects with list of files for each project
        Dictionary<Project, Dictionary<string, string>> projects = new Dictionary<Project, Dictionary<string, string>>();
        for (int i = 1; i < _dte.Solution.Projects.Count + 1; i++)
        {
          Project project = _dte.Solution.Projects.Item(i);
          // solution folders are "projects" - drill down for nested real projects
          if (IsSolutionFolder(project.Kind))
            IterateSubProjects(project, projects);
          else
            IterateProjectItems(project, projects);
        }

        Dictionary<Project, List<string>> filesProjects = new Dictionary<Project, List<string>>();
        // iterate over pending changes, and handle relevant only
        foreach (PendingChange change in PendingCheckin.PendingChanges.CheckedPendingChanges)
        {
          Trace.TraceInformation("Evaluate change {1} {0} {2}", change.LocalItem, change.ChangeType, change.ItemType);
          // ignore folders
          if (change.ItemType == ItemType.Folder)
            continue;
          // handle only new files or edits
          if (!change.IsAdd && !change.IsEdit)
            continue;
          // ignore non-C# files
          if (String.Compare(Path.GetExtension(change.LocalItem), ".cs", true) != 0)
            continue;
          // TODO: handle *.g.cs && *.designer.cs
          // determine VS project for the file and add to project list
          Project project = FindProject(projects, change.LocalItem);
          Trace.TraceInformation("Evaluate project {0} for {1}", project, change.LocalItem);
          // if some of the files checked out not in current solution projects, we should fail
          if (project == null)
          {
            _externalFiles.Add(change.LocalItem);
            continue;
          }
          List<string> files;
          // make connection between project and files
          if (filesProjects.ContainsKey(project))
            files = filesProjects[project];
          else
          {
            files = new List<string>();
            filesProjects.Add(project, files);
          }
          files.Add(change.LocalItem);
          _fileProjectMap[change.LocalItem] = project;
        }
        // at least some files being checked in are not in projects
        if (_externalFiles.Count > 0)
          return new PolicyFailure[] { new PolicyFailure("Some files were not found in current solution and cannot be analyzed (double click for details)", this) };
        // No C# files found in check in set, then we are good
        if (filesProjects.Count == 0)
          return new PolicyFailure[] { };
        // have some files to process - run analysis (will fill violations list)
        if (!ExecuteSourceAnalysis(_dte.Solution.FileName, filesProjects))
        {
          if (_violations != null && _violations.Count > 0)
          {
            List<PolicyFailure> results = new List<PolicyFailure>(_violations.Count);
            // prepare list of violations to show in policy window
            foreach (string projectKey in _violations.Keys)
            {
              List<ViolationEventArgs> list = _violations[projectKey];
              foreach (ViolationEventArgs issue in list)
              {
                string message = issue.Violation.Rule.CheckId + ": " + issue.Message;
                results.Add(new PolicyFailure(String.Format("{0} {1} {2}{3}{4}{5}", projectKey, message, ErrorFilePrefix, GetFileName(issue), ErrorFileLinePrefix, issue.LineNumber), this));
              }
            }

            // clean previous violations in VS errors window
            _errorsProvider.Tasks.Clear();
            // need to keep list of tasks alive here
            _errorTasks = new List<ErrorTask>();

            // get current solution service
            IVsSolution solution = Microsoft.VisualStudio.Shell.Package.GetGlobalService(typeof(IVsSolution)) as IVsSolution;
            // get build manager service, try to unadvise previous sink, advise new sink for build events
            IVsSolutionBuildManager buildManager = Microsoft.VisualStudio.Shell.Package.GetGlobalService(typeof(IVsSolutionBuildManager)) as IVsSolutionBuildManager;
            if (_solutionBuildEventsCookie != 0)
            {
              buildManager.UnadviseUpdateSolutionEvents(_solutionBuildEventsCookie);
              _solutionBuildEventsCookie = 0;
            }
            buildManager.AdviseUpdateSolutionEvents(this, out _solutionBuildEventsCookie);

            // iterate over violations and add them to error list
            foreach (string projectKey in _violations.Keys)
            {
              List<ViolationEventArgs> list = _violations[projectKey];
              foreach (ViolationEventArgs issue in list)
              {
                ErrorTask errorTask = CreateErrorTask(issue, _fileProjectMap, solution);
                errorTask.Navigate += ErrorTaskNavigate;
                _errorsProvider.Tasks.Add(errorTask);
                _errorTasks.Add(errorTask);
              }
            }
            // might want to focus on errors list
            //_errorsProvider.Show();

            return results.ToArray();
          }


          return new PolicyFailure[] { new PolicyFailure("Source Analysis failed (double click for details)", this) };
        }
        return new PolicyFailure[] { };
      }
      catch (Exception ex)
      {
        Trace.TraceError("Evaluate error {0} {1}", ex.Message, ex.StackTrace);
        throw ex;
      }
    }

    private static ErrorTask CreateErrorTask(ViolationEventArgs issue, Dictionary<string, Project> fileProjectMap, IVsSolution solution)
    {
      string fileName = GetFileName(issue);
      Project project = null;
      if (fileProjectMap.ContainsKey(fileName))
      {
        project = fileProjectMap[fileName];
      }

      IVsHierarchy hierarchy;
      solution.GetProjectOfUniqueName(project.UniqueName, out hierarchy);

      ErrorTask errorTask = new ErrorTask();
      errorTask.CanDelete = true;
      errorTask.Category = TaskCategory.CodeSense;
      errorTask.ErrorCategory = TaskErrorCategory.Error;
      errorTask.HierarchyItem = hierarchy;
      errorTask.Document = fileName;
      errorTask.Line = issue.LineNumber - 1;
      errorTask.Column = 0;
      errorTask.Text = issue.Violation.Rule.CheckId + ": " + issue.Message;

      return errorTask;
    }

    /// <summary>
    /// Gets file name out of StyleCop violation
    /// </summary>
    private static string GetFileName(ViolationEventArgs violation)
    {
      if ((violation.SourceCode != null && violation.SourceCode.Path != null) && violation.SourceCode.Path.Length > 0)
        return violation.SourceCode.Path;
      if ((violation.Element != null && violation.Element.Document != null) && (violation.Element.Document.SourceCode != null && violation.Element.Document.SourceCode.Path != null))
        return violation.Element.Document.SourceCode.Path;
      return "None";
    }

    /// <summary>
    /// Navigation handler for StyleCop errors in Error list
    /// </summary>
    private void ErrorTaskNavigate(object sender, EventArgs e)
    {
      ErrorTask task = sender as ErrorTask;
      Debug.Assert(task != null);
      task.Line += 1;
      _errorsProvider.Navigate(task, new Guid(EnvDTE.Constants.vsViewKindCode));
      task.Line -= 1;
    }

    /// <summary>
    /// Holds list of files not found in currently loaded projects
    /// </summary >
    [NonSerialized]
    private List<string> _externalFiles;

    /// <summary>
    /// Maps source C# file to its parent project
    /// </summary>
    [NonSerialized]
    private Dictionary<string, Project> _fileProjectMap;
    /// <summary>
    /// List of violations filled by source analysis engine
    /// </summary>
    [NonSerialized]
    private Dictionary<string, List<ViolationEventArgs>> _violations;

    private static string CreateSettingsFile(string solutionPath, string settingsContents)
    {
      // try to create file next to solution (<solution file name>.sln.StyleCop)
      string settingsFilePath = solutionPath + ".StyleCop";
      // if settings file exists, get rid of it
      try
      {
        if (File.Exists(settingsFilePath))
        {
          if ((File.GetAttributes(settingsFilePath) & FileAttributes.ReadOnly) == FileAttributes.ReadOnly)
            File.SetAttributes(settingsFilePath, FileAttributes.Normal);
          File.Delete(settingsFilePath);
        }
      }
      catch (Exception ex)
      {
        Trace.TraceError("Error removing solution definitions file {0} \r\n {1}", ex.Message, ex.StackTrace);
        settingsFilePath = Path.GetTempFileName();
      }
      // write settings contents into file
      Debug.Assert(!String.IsNullOrEmpty(settingsContents));
      using (StreamWriter writer = new StreamWriter(File.Open(settingsFilePath, FileMode.Create)))
      {
        writer.Write(settingsContents);
        writer.Close();
      }

      return settingsFilePath;
    }

    private static bool IsTemporarySettingsFile(string solutionPath, string settingsFilePath)
    {
      return !settingsFilePath.StartsWith(solutionPath, StringComparison.CurrentCultureIgnoreCase);
    }

    private static void DeleteFile(string filePath)
    {
      try
      {
        File.Delete(filePath);
      }
      catch (Exception ex)
      {
        Trace.TraceError("Exception while deleting file {0} {1} {2}", filePath, ex.Message, ex.StackTrace);
      }
    }

    /// <summary>
    /// Executes source analysis, given list of projects with list of files to analyze for each 
    /// </summary>
    /// <param name="projects"></param>
    private bool ExecuteSourceAnalysis(string solutionPath, Dictionary<Project, List<string>> projects)
    {
      string settingsFilePath = CreateSettingsFile(solutionPath, _sourceCodeAnalysisSettings);

      StyleCopConsole console = new StyleCopConsole(settingsFilePath,
          false, null, new List<string>(), true);
      // TODO: fix constants later?
      Microsoft.StyleCop.Configuration configuration = new Microsoft.StyleCop.Configuration(new string[] { });// project.Defines);
      // initialize all projects to analyze
      List<CodeProject> codeProjects = new List<CodeProject>();
      foreach (Project project in projects.Keys)
      {
        Microsoft.StyleCop.CodeProject codeProject = new Microsoft.StyleCop.CodeProject(project.GetHashCode(), project.Name, configuration);
        List<string> files = projects[project];
        foreach (string file in files)
          console.Core.Environment.AddSourceCode(codeProject, file, null);
        codeProjects.Add(codeProject);
      }

      // run source analysis (events will fill violations list)
      try
      {
        console.OutputGenerated += new EventHandler<OutputEventArgs>(OnOutputGenerated);
        console.ViolationEncountered += new EventHandler<ViolationEventArgs>(OnViolationEncountered);
        console.Start(codeProjects.ToArray(), true);
        return (_violations.Count == 0);
      }
      finally
      {
        if (IsTemporarySettingsFile(solutionPath, settingsFilePath))
          DeleteFile(settingsFilePath);
        console.OutputGenerated -= new EventHandler<OutputEventArgs>(OnOutputGenerated);
        console.ViolationEncountered -= new EventHandler<ViolationEventArgs>(OnViolationEncountered);
      }
    }

    private void OnOutputGenerated(object sender, OutputEventArgs e)
    {
      Trace.TraceInformation("OnOutputGenerated {0}", e.Output.Trim());
    }

    private void OnViolationEncountered(object sender, ViolationEventArgs e)
    {
      string file = GetFileName(e);
      string message = e.Violation.Rule.CheckId + ": " + e.Message;
      Trace.TraceInformation("OnViolationEncountered {0}, {1}, {2}", file, e.LineNumber, message);
      // put the result into dictionary (for details)
      Debug.Assert(_fileProjectMap.ContainsKey(file));
      Project project = _fileProjectMap[file];
      if (!_violations.ContainsKey(project.UniqueName))
        _violations.Add(project.UniqueName, new List<ViolationEventArgs>());
      List<ViolationEventArgs> issues = _violations[project.UniqueName];
      issues.Add(e);
    }

    #region IVsUpdateSolutionEvents Members

    public int OnActiveProjectCfgChange(IVsHierarchy pIVsHierarchy)
    {
      return VSConstants.S_OK;
    }

    public int UpdateSolution_Begin(ref int pfCancelUpdate)
    {
      if (_errorsProvider != null)
      {
        _errorsProvider.Tasks.Clear();
        _errorTasks.Clear();
      }
      return VSConstants.S_OK;
    }

    public int UpdateSolution_Cancel()
    {
      if (_errorsProvider != null)
      {
        _errorsProvider.Tasks.Clear();
        _errorTasks.Clear();
      }
      return VSConstants.S_OK;
    }

    public int UpdateSolution_Done(int fSucceeded, int fModified, int fCancelCommand)
    {
      return VSConstants.S_OK;
    }

    public int UpdateSolution_StartUpdate(ref int pfCancelUpdate)
    {
      return VSConstants.S_OK;
    }

    #endregion
  }
}
